#include <PalmOS.h>


//we flip the 0x40000000 bit to mark alarms as ours as all real palmos pointers have that as low

#define DB_ID_MASK_OURS			0x40000000ul

//verified 1.0 to 3.1 - our relevant OS range
struct AlarmEntry {
	UInt32 ref;					//userData
	UInt32 at;					// trigger time (0 if entry is empty)
	LocalID LID;
	UInt8 cardNo		: 1;	//yes, one bit
	UInt8 quiet			: 1;
	UInt8 triggered		: 1;
	UInt8 notifed		: 1;
};

struct AlarmTable {
	UInt16 nEntres;
	struct AlarmEntry entry[];
};

struct PalmOsGlobals {		//based at 2 to avoid gcc issues
	UInt8 whatever1[0x37e];
	MemHandle alarmGlobals;
};

struct ProcAlarmCallbackInfo {
	AlmAlarmProcPtr proc;
	UInt32 ref, at;
};

Err pAlmSetAlarm(Err (*oldTrap)(UInt16 cardNo, LocalID dbID, UInt32 ref, UInt32 alarmSeconds, Boolean quiet), void *retAddr, UInt16 cardNo, LocalID dbID, UInt32 ref, UInt32 alarmSeconds, Boolean quiet)
{
	if (cardNo == almProcAlarmCardNo) {
	
		cardNo = 1;
		dbID ^= DB_ID_MASK_OURS;
	}
	
	return oldTrap(cardNo, dbID, ref, alarmSeconds, quiet);
}

UInt32 pAlmGetAlarm(UInt32 (*oldTrap)(UInt16 cardNo, LocalID dbID, UInt32 *refP), void *retAddr, UInt16 cardNo, LocalID dbID, UInt32 *refP)
{

	if (cardNo == almProcAlarmCardNo) {
	
		cardNo = 1;
		dbID ^= DB_ID_MASK_OURS;
	}
	
	return oldTrap(cardNo, dbID, refP);
}

Err pSysAppLaunch(Err (*oldTrap)(UInt16 cardNo, LocalID dbID, UInt16 launchFlags, UInt16 cmd, MemPtr cmdPBP, UInt32 *resultP), void *retAddr, UInt16 cardNo, LocalID dbID, UInt16 launchFlags, UInt16 cmd, MemPtr cmdPBP, UInt32 *resultP)
{
	Err ret;
	
	char x[33];
	
	
	if (cardNo == 1 && (dbID & DB_ID_MASK_OURS) && cmd == sysAppLaunchCmdAlarmTriggered) {
		
		struct SysAlarmTriggeredParamType *p = (struct SysAlarmTriggeredParamType*)cmdPBP;
		
		((AlmAlarmProcPtr)(dbID ^ DB_ID_MASK_OURS))(almProcCmdTriggered, p);
		p->purgeAlarm = true;
		
		ret = errNone;
	}
	else
		ret = oldTrap(cardNo, dbID, launchFlags, cmd, cmdPBP, resultP);
	
	return ret;
}

Err pSysBroadcastActionCode(Err (*oldTrap)(UInt16 cmd, MemPtr cmdPBP), void *retAddr, UInt16 cmd, MemPtr cmdPBP)
{
	if (cmd == sysAppLaunchCmdTimeChange) {
		
		struct PalmOsGlobals *pg = (struct PalmOsGlobals*)2;
		MemHandle mh = pg->alarmGlobals;
		struct AlarmTable *tab;
		UInt16 i, j;
		
		//we need to cancel each procedural alarm, AND then fire a callback
		
		if (mh){
			
			tab = MemHandleLock(mh);
			
			if (tab->nEntres) {
			
				struct ProcAlarmCallbackInfo *cbi = MemPtrNew(sizeof(struct ProcAlarmCallbackInfo) * tab->nEntres);
				struct SysAlarmTriggeredParamType tp;
				
				//firt we'll find them all and record them
				for (i = 0, j = 0; i < tab->nEntres; i++) {
					
					//we only care about active procedural alarm entries
					if (!tab->entry[i].at || !tab->entry[i].cardNo || !(tab->entry[i].LID & DB_ID_MASK_OURS))
						continue;
					
					cbi[j].proc = (AlmAlarmProcPtr)(tab->entry[i].LID ^ DB_ID_MASK_OURS);
					cbi[j].ref = tab->entry[i].ref;
					cbi[j].at = tab->entry[i].at;
					j++;
				}
				
				//then we'll cancel them all
				for (i = 0; i < j; i++)
					(void)AlmSetAlarm(almProcAlarmCardNo, (LocalID)cbi[i].proc, cbi[i].ref, 0, false);
				
				//then we'll notify them all
				for (i = 0; i < j; i++) {
				
					tp.ref = cbi[i].ref;
					tp.alarmSeconds = cbi[i].at;
					tp.purgeAlarm = true;
					
					cbi[i].proc(almProcCmdReschedule, &tp);
				}
				
				MemChunkFree(cbi);
			}
			MemHandleUnlock(mh);
		}
	}
	
	return oldTrap(cmd, cmdPBP);
}

#ifdef __GNUC__
	
	static void pAlmSetAlarmWrapper(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		//also gcc gets too clever with tailcalls, so to avoid them, we provide a proper stack
		//sadly this means we need to create pointless stack frames
		//c'est la vie
		
		asm volatile(
			"	move.l  #0xaaaaaaaa, %d0			\n\t"		//old trap
			"	move.w  18(%sp), -(%sp)				\n\t"
			"	move.l  16(%sp), -(%sp)				\n\t"
			"	move.l  16(%sp), -(%sp)				\n\t"
			"	move.l  16(%sp), -(%sp)				\n\t"
			"	move.w  18(%sp), -(%sp)				\n\t"
			"	clr.l   -(%sp)						\n\t"
			"	move.l  %d0, -(%sp)					\n\t"
			"	bsr    pAlmSetAlarm					\n\t"
			"	lea    24(%sp), %sp					\n\t"
		);
	}
	
	static void pAlmGetAlarmWrapper(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		//also gcc gets too clever with tailcalls, so to avoid them, we provide a proper stack
		//sadly this means we need to create pointless stack frames
		//c'est la vie
		
		asm volatile(
			"	move.l  #0xaaaaaaaa, %d0			\n\t"		//old trap
			"	move.l  10(%sp), -(%sp)				\n\t"
			"	move.l  10(%sp), -(%sp)				\n\t"
			"	move.w  12(%sp), -(%sp)				\n\t"
			"	clr.l   -(%sp)						\n\t"
			"	move.l  %d0, -(%sp)					\n\t"
			"	bsr    pAlmGetAlarm					\n\t"
			"	lea    18(%sp), %sp					\n\t"
		);
	}
	
	static void pSysAppLaunchWrapper(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		//also gcc gets too clever with tailcalls, so to avoid them, we provide a proper stack
		//sadly this means we need to create pointless stack frames
		//c'est la vie
		
		asm volatile(
			"	move.l  #0xaaaaaaaa, %d0			\n\t"		//old trap
			"	move.l  18(%sp), -(%sp)				\n\t"
			"	move.l  18(%sp), -(%sp)				\n\t"
			"	move.l  18(%sp), -(%sp)				\n\t"	//copy two at once
			"	move.l  18(%sp), -(%sp)				\n\t"
			"	move.w  20(%sp), -(%sp)				\n\t"
			"	clr.l   -(%sp)						\n\t"
			"	move.l  %d0, -(%sp)					\n\t"
			"	bsr    pSysAppLaunch				\n\t"
			"	lea    26(%sp), %sp					\n\t"
		);
	}
	
	static void pSysBroadcastActionCodeWrapper(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		//also gcc gets too clever with tailcalls, so to avoid them, we provide a proper stack
		//sadly this means we need to create pointless stack frames
		//c'est la vie
		
		asm volatile(
			"	move.l  #0xaaaaaaaa, %d0			\n\t"		//old trap
			"	move.l  6(%sp), -(%sp)				\n\t"
			"	move.w  8(%sp), -(%sp)				\n\t"
			"	clr.l   -(%sp)						\n\t"
			"	move.l  %d0, -(%sp)					\n\t"
			"	bsr    pSysBroadcastActionCode		\n\t"
			"	lea    14(%sp), %sp					\n\t"
		);
	}
	
#else
	
	asm void pAlmSetAlarmWrapper(void)
	{
		move.l #0xaaaaaaaa, -(sp)			//old trap
		jsr    pAlmSetAlarm
		lea    4(sp), sp
		rts
	}
	
	asm void pAlmGetAlarmWrapper(void)
	{
		move.l #0xaaaaaaaa, -(sp)			//old trap
		jsr    pAlmGetAlarm
		lea    4(sp), sp
		rts
	}
	
	asm void pSysAppLaunchWrapper(void)
	{
		move.l #0xaaaaaaaa, -(sp)			//old trap
		jsr    pSysAppLaunch
		lea    4(sp), sp
		rts
	}
	
	asm void pSysBroadcastActionCodeWrapper(void)
	{
		move.l #0xaaaaaaaa, -(sp)			//old trap
		jsr    pSysBroadcastActionCode
		lea    4(sp), sp
		rts
	}
	
#endif

void __Startup__(void)
{
	UInt32 ver;
		
	if (errNone == FtrGet(sysFtrCreator, sysFtrNumROMVersion, &ver) && ver < sysMakeROMVersion(3, 2, 0, 0, 0)) {
		
		MemSemaphoreReserve(true);
		*(volatile void**)(((UInt16*)pAlmSetAlarmWrapper) + 1) = SysGetTrapAddress(sysTrapAlmSetAlarm);
		*(volatile void**)(((UInt16*)pAlmGetAlarmWrapper) + 1) = SysGetTrapAddress(sysTrapAlmGetAlarm);
		*(volatile void**)(((UInt16*)pSysAppLaunchWrapper) + 1) = SysGetTrapAddress(sysTrapSysAppLaunch);
		*(volatile void**)(((UInt16*)pSysBroadcastActionCodeWrapper) + 1) = SysGetTrapAddress(sysTrapSysBroadcastActionCode);
		MemSemaphoreRelease(true);
		
		SysSetTrapAddress(sysTrapAlmSetAlarm,pAlmSetAlarmWrapper);
		SysSetTrapAddress(sysTrapAlmGetAlarm, pAlmGetAlarmWrapper);
		SysSetTrapAddress(sysTrapSysAppLaunch, pSysAppLaunchWrapper);
		SysSetTrapAddress(sysTrapSysBroadcastActionCode, pSysBroadcastActionCodeWrapper);
		
		MemHandleLock(DmGet1Resource('extn', 11));
	}
}
