#include <PalmOS.h>


#define HEAP_ID_STORAGE						1
#define OWNER_ID_FTR_ALLOCATED_CHUNKS		3

static Err pFtrPtrNew(UInt32 creator, UInt16 featureNum, UInt32 size, void **newPtrP)
{
	MemHandle mh;
	void *ret;
	Err e;
	
	if (!size)
		return memErrInvalidParam;
	
	//XXX: on 1.0 and 2.0 might be worth trying ALL storage heaps
	mh = MemChunkNew(HEAP_ID_STORAGE, size, memNewChunkFlagAllowLarge | OWNER_ID_FTR_ALLOCATED_CHUNKS);
	if (!mh)
		return memErrNotEnoughSpace;
	
	ret = MemHandleLock(mh);
	
	e = FtrSet(creator, featureNum, (UInt32)ret);
	if (e != errNone) {
		
		MemChunkFree(ret);
		return e;
	}
	
	if (newPtrP)
		*newPtrP = ret;
	
	return errNone;
}

Err pFtrPtrFree(UInt32 creator, UInt16 featureNum)
{
	void* ptr;
	Err e;
	
	e = FtrGet(creator, featureNum, (UInt32*)&ptr);
	if (e != errNone)
		return e;

	e = MemChunkFree(ptr);
	if (e != errNone)
		return e;
	
	FtrUnregister(creator, featureNum);
	return errNone;
}

Err pFtrPtrResize (UInt32 creator, UInt16 featureNum, UInt32 newSize, void **newPtrP)
{
	MemHandle mh;
	void* ptr;
	Err e;
	
	if (!newSize)
		return memErrInvalidParam;
	
	e = FtrGet(creator, featureNum, (UInt32*)&ptr);
	if (e != errNone)
		return e;
	
	mh = MemPtrRecoverHandle(ptr);
	MemHandleUnlock(mh);
	e = MemHandleResize(mh, newSize);
	ptr = MemHandleLock(mh);
	
	if (e != errNone)
		return e;
	
	e = FtrSet(creator, featureNum, (UInt32)ptr);
	if (e != errNone) {
		
		FtrUnregister(creator, featureNum);
		MemChunkFree(ptr);
		return e;
	}
	
	if (newPtrP)
		*newPtrP = ptr;
	
	return errNone;
}


void __Startup__(void)
{
	UInt32 ver;
	
	if (errNone == FtrGet(sysFtrCreator, sysFtrNumROMVersion, &ver) && ver < sysMakeROMVersion(3, 1, 0, 0, 0)) {
		
		SysSetTrapAddress(sysTrapFtrPtrNew, pFtrPtrNew);
		SysSetTrapAddress(sysTrapFtrPtrResize, pFtrPtrResize);
		SysSetTrapAddress(sysTrapFtrPtrFree, pFtrPtrFree);
		
		MemHandleLock(DmGet1Resource('extn', 10));
	}
}
