#include <PalmOS.h>


struct SysLibTblEntryTypeV10 {		//keep this in mind - this is what PalmOS 1.0 uses!!!!!  
	MemPtr *dispatchTblP;
	void *globalsP;
};

//PalmOS 1.0's SysLibInstall does not clear "globalsP" which some libraries expect...
//Luckily, the param to SysLibInstall is only used there at install time, so we can resolve this


/*
static Err pSysLibEntryProcPtrInterposer(UInt16 refNum, struct SysLibTblEntryTypeV10 *entryP)
00000000: 206F 0006          movea.l   6(a7),a0					//a0 = param_entryP
00000004: 42A8 0004          clr.l     4(a0)					//param_entryP->globalsP = 0
00000008: 207C 1234 5678     movea.l   #0x12345678,a0			//a0 = real_entry_pt
0000000E: 4ED0               jmp       (a0)						//go there
*/

//why this and just not use a feature?
//things go bad when we use features inside library loading code due to OS bugs and nonreentrancy.
//this is also simpler :)

Err pSysLibLoad(UInt32 type, UInt32 crid, UInt16 *libRefP)
{
	struct SysLibTblEntryTypeV10 *lte;
	DmSearchStateType ss;
	UInt16 cardNo;
	MemHandle mh;
	DmOpenRef db;
	LocalID LID;
	Err e;
	
	if (!libRefP)
		return sysErrParamErr;
	*libRefP = sysInvalidRefNum;

	e = DmGetNextDatabaseByTypeCreator(true, &ss, type, crid, true, &cardNo, &LID);
	if (e  != errNone || !LID)
		return sysErrLibNotFound;
	

	db = DmOpenDatabase(cardNo, LID, dmModeReadOnly | dmModeLeaveOpen);
	if (!db)
		return DmGetLastErr();
	
	mh = DmGet1Resource(sysResTLibrary, 0);
	if (!mh)
		e = sysErrLibNotFound;
	else {
		
		UInt32 addr = (UInt32)MemHandleLock(mh);
		UInt16 code[] = {0x206F, 0x0006, 0x42a8, 0x0004, 0x207c, addr >> 16, addr, 0x4ed0};
		
		e = SysLibInstall((SysLibEntryProcPtr)code, libRefP);
		if (e != errNone) {
			
			MemHandleUnlock(mh);
			DmReleaseResource(mh);
		}
		else {
			
			(void)DmDatabaseProtect(cardNo, LID, true);
		}
	}
	
	DmCloseDatabase(db);
	return e;
}

#ifdef __GNUC__

	static void pSysLibLoad_pPsrClose(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		
		asm volatile(
			"	move.l  (%sp), %d0			\n\t"
			"	subi.l  #0xAAAAAAAA, %d0	\n\t"	//start addr of System.prc('boot', 10102)
			"	bmi.s   not_rom				\n\t"
			"	subi.l  #0xAAAAAAAA, %d0	\n\t"	//length of System.prc('boot', 10102)
			"	bpl.s   not_rom				\n\t"
			"	movea.l #0xAAAAAAAA, %a0	\n\t"	//old trap
			"	jmp     (%a0)				\n\t"
			"not_rom:						\n\t"
			"	bra     pSysLibLoad			\n\t"
		);
	}

#else

	asm static void pSysLibLoad_pPsrClose(void)
	{
		move.l  (sp), d0
		subi.l  #0xAAAAAAAA, d0	//start addr of System.prc('boot', 10102)
		bmi     not_rom
		subi.l  #0xAAAAAAAA, d0	//length of System.prc('boot', 10102)
		bpl     not_rom
		movea.l #0xAAAAAAAA, a0	//old trap
		jmp     (a0)
	not_rom:
		jmp     pSysLibLoad
	}

#endif

void __Startup__(void)
{
	UInt32 ver;
	
	if (errNone == FtrGet(sysFtrCreator, sysFtrNumROMVersion, &ver) && ver < sysMakeROMVersion(2, 0, 0, 0, 0)) {
		
		//PalmOS 1.0 only:
		// We want to allow existing code to use the old code allocated to sysTrapSysLibLoad.
		// That would be sysTrapPsrClose. Luckily, the only code that does is in System.prc, ('boot', 10102) resource
		// we check for return address in there to decide what to do.
		// I hope nobody else patches this func after us, else this logic will break
		
		DmOpenRef db = DmOpenDatabaseByTypeCreator(sysFileTSystem, sysFileCSystem, dmModeReadOnly);
		MemHandle mh = DmGet1Resource(sysResTBootCode, 10102);
		UInt32 sysCodeSize = MemHandleSize(mh);
		void *sysCodePtr = MemHandleLock(mh);
		
		MemSemaphoreReserve(true);
		*(UInt32*)(((UInt16*)pSysLibLoad_pPsrClose) + 2) = (UInt32)sysCodePtr;
		*(UInt32*)(((UInt16*)pSysLibLoad_pPsrClose) + 6) = sysCodeSize;
		*(UInt32*)(((UInt16*)pSysLibLoad_pPsrClose) + 10) = (UInt32)SysGetTrapAddress(sysTrapSysLibLoad);
		MemSemaphoreRelease(true);
		
		MemHandleUnlock(mh);
		DmReleaseResource(mh);
		DmCloseDatabase(db);
		
		SysSetTrapAddress(sysTrapSysLibLoad, pSysLibLoad_pPsrClose);
		
		MemHandleLock(DmGet1Resource('extn', 5));
	}
}

