#include <SerialMgrOld.h>
#include "bbUart.h"

struct Uart68K {
	volatile UInt16 ustcnt;
	volatile UInt16 ubaud;
	volatile UInt16 urx;
	volatile UInt16 utx;
	volatile UInt16 umisc;
	volatile UInt16 nipr;	//ez+
	UInt16 rfu[2];
};

struct Gpio68K {
	volatile UInt8 direction;
	volatile UInt8 data;
	volatile UInt8 pullup;
	volatile UInt8 select;
};


//this is inverted uart, so start bit is high and stop bit is low
static asm UInt16 bbUArtTxThenRxLo(struct UartState *state, Int16 tx, void *dst, UInt16 maxBytes, UInt32 maxWait)	//return num bytes RXed. do not clal with "maxBytes == 0"!!!
{
	//get params
	move.l  4(sp), a1	//state
	move.w  8(sp), d0	//tx byte				xxxx
	move.l  10(sp), a0	//dst
	move.w  14(sp), d1	//maxBytes
	move.l  16(sp), d2	//max wait
	
	//get some more regs to play with
	movem.l d3-d6,-(sp)
	
	//get some stuff out of "state"
	move.b  8(a1), d6   //txBit					xxxxx
	move.b  9(a1), d3   //rxBit
	move.l  4(a1), a1	//state->gpio
	adda.w  #1, a1		//&state->gpio->data
	
	
	//if we do not need to TX, skip it
	move.w  d0, d0
	bmi     rx_start
	
	//tx  abyte
	ori.w   #0xff00, d0
	add.w   d0, d0
	
	move.w  #8, d5			//9 iterations -> no stop bits
txbitloop:					//2094
	lsr.w   #1, d0			//8
	bcc.s   send_zero		//8 or 10
	bcs.s   send_one		//8 or 10
	
send_zero:
	bcs.s   bitloop			//8			dummy not-taken branch to keep lengths same
	bset    d6, (a1)		//12
	bra.s   sent			//10
	
send_one:
	bclr    d6, (a1)		//12
	bra.s   sent			//10
	
	nop						//needed fot the bra.s above to exist
	
sent:
	nop						//4
	move.w  #400, d4		//8
delaytx:					//(d4 + 1) * 10 + 14 = 2024
	dbra.s  d4, delaytx		//10 if looping, 14 if not

	dbra.s  d5, txbitloop	//10 if looping
	
	bclr    d6, (a1)		//put line back to idle since we never sent the stop bit
	
	//after TX first time we need to wait for line to go idle since the chip might have just turned on the transciever
wait_xcvr_onstart:

	//rx stuff
	
rx_start:
	
	//sanity check
	subi.w  #1, d1
	bmi.s   out
	
rx_byte_loop:
	
	//wait for a start bit
wait_start:
	btst    d3, (a1)
	bne.s   start_byte_rx
	moveq   #1, d4
	sub.l   d4, d2
	bne.s   wait_start
	bra     out
	
start_byte_rx:
	moveq   #7, d5			//4
	move.w  #550, d4		//8
one_and_halfdelay:			//(d4 + 1) * 10 + 14 = ?
	dbra.s  d4, one_and_halfdelay

bitloop:					//2194
	btst    d3, (a1)		//4
	beq.s   got_one			//10 if taken, 8 if not
	bne.s   got_zero		//10 if taken, not possible to not

got_one:
	bne.s   delay			//8		dummy not-taken branch to keep lengths same
	lsr     #1, d0			//8
	ori.b   #0x80, d0		//8
	bra.s   got_bit			//10

got_zero:
	lsr     #1, d0			//8
	ori.b   #0x00, d0		//8
	bra.s   got_bit			//10
	
	nop						//needed for the bra.s above to exist
	
got_bit:
	move.w  #409, d4		//8
delay:						//(d4 + 1) * 10 + 14 = 2114
	dbra.s  d4, delay

	nop						//4

	dbra.s  d5, bitloop		//10 when taken, 14 when not

got_byte:
		
	//verify framing and skip if not valid
	btst    d3, (a1)		//4
	bne.s   rx_byte_loop	//10 when taken, 8 when not

	//store
	move.b  d0, (a0)+		//8
	dbra.s  d1, rx_byte_loop//10 when taken, 14 when not

out:
	//restore regs
	movem.l (sp)+,d3-d6
	
	//calc return value
	suba.l  10(sp), a0
	move.w  a0, d0
	rts
	
out_error:
	moveq   #0, d0
	rts
}

UInt16 bbUArtTxThenRx(struct UartState *state, Int16 tx, void *dst, UInt16 maxBytes, UInt32 maxWait)
{
	UInt16 irqState, ret;
	
	irqState = SysDisableInts();
	ret = bbUArtTxThenRxLo(state, tx, dst, maxBytes, maxWait);
	SysRestoreStatus(irqState);
	
	return ret;
}

Boolean bbUartOpen(struct UartState *state, enum UartOption uo)
{
	UInt16 serLibRef;
	Err e;
	
	e = SysLibFind("Serial Library", &state->serLibRef);
	if (e != errNone)
		return false;

	e = SerOpen(state->serLibRef, serPortDefault, 57600 /* does not matter, but high enough to require handshaking (in case we want to check for that) */);
	if (e != errNone)
		return false;

	switch (uo) {
		case Uart328:
			state->uart = (struct Uart68K*)0xfffff900ul;
			state->gpio = (struct Gpio68K*)0xfffff430ul;
			state->txBit = 0;
			state->rxBit = 1;
			break;
		
		case UartEZ:
		case UartNumberOne:
			state->uart = (struct Uart68K*)0xfffff900ul;
			state->gpio = (struct Gpio68K*)0xfffff420ul;
			state->txBit = 5;
			state->rxBit = 4;
			break;
		
		case UartNumberTwo:
			state->uart = (struct Uart68K*)0xfffff910ul;
			state->gpio = (struct Gpio68K*)0xfffff438ul;
			state->txBit = 5;
			state->rxBit = 4;
			break;
		
		default:
			return false;
	}
	
	state->gpio->data &=~ (1 << state->txBit);	//idle low
	state->gpio->direction = (state->gpio->direction &~ (1 << state->rxBit)) | (1 << state->txBit);
	state->gpio->select |= (1 << state->rxBit) | (1 << state->txBit);
	
	return true;
}

void bbUartClose(struct UartState *state)
{
	state->gpio->select &=~ ((1 << state->rxBit) | (1 << state->txBit));
	SerClose(state->serLibRef);
}

enum UartOption portPrvFind(void)
{
	struct Uart68K *uartHw = (struct Uart68K*)0xfffff900ul;
	Boolean candidate[2] = {true, true};
	UInt16 serLibRef;
	UInt32 cpu;
	UInt8 i;
	Err e;
	
	
	if (errNone != FtrGet(sysFtrCreator, sysFtrNumProcessorID, &cpu))
		return UartUnknown;
	
	switch (cpu >> 16) {
		case sysFtrNumProcessor328 >> 16:
			return Uart328;
		
		case sysFtrNumProcessorEZ >> 16:
			return UartEZ;
		
		case sysFtrNumProcessorVZ >> 16:
		case sysFtrNumProcessorSuperVZ >> 16:
			break;
		
		default:
			return UartUnknown;
	}
	
	//read state before opening port
	for (i = 0; i < 2; i++)
		candidate[i] &= !(uartHw[i].ustcnt >> 15);
	
	//open port
	e = SysLibFind("Serial Library", &serLibRef);
	if (e != errNone)
		return UartUnknown;

	e = SerOpen(serLibRef, serPortDefault, 57600 /* does not matter, but high enough to require handshaking (in case we want to check for that) */);
	if (e != errNone)
		return UartUnknown;
	
	//read state after opening port
	for (i = 0; i < 2; i++)
		candidate[i] &= uartHw[i].ustcnt >> 15;
	
	//close port
	(void)SerClose(serLibRef);
	
	//decide
	if (candidate[0] && candidate[1])
		return UartUnknown;
	if (candidate[0])
		return UartNumberOne;
	if (candidate[1])
		return UartNumberTwo;
	return UartUnknown;
}

