#include <ExpansionMgr.h>
#include <SlotDrvrLib.h>
#include <VFSMgr.h>
#include <PalmOS.h>
#include "bbUart.h"
#include "update.h"
#include "iface.h"




#define PBAR_LEFT		3
#define PBAR_TOP		115
#define PBAR_WIDTH		150
#define PBAR_HEIGHT		16


//this nonsense is needed to avoid strings ending up in ".data" somehow
static const char mStrTaskFailedSuccessfully[] = "Task failed successfully";
static const char mStrAnErrorOccured[] = "An error occured: ";
static const char mStrOccured[] = " occured";
static const char mStrError0x[] = "Error 0x";




static void uiPrvPbarErase(void)
{
	RectangleType r;
	
	r.topLeft.x = PBAR_LEFT;
	r.topLeft.y = PBAR_TOP;
	r.extent.x = PBAR_WIDTH;
	r.extent.y = PBAR_HEIGHT;
	
	WinEraseRectangle(&r, 0);
}

static void uiPrvPbarDraw(UInt16 num, UInt16 of)
{
	RectangleType r, e;
	
	r.topLeft.x = PBAR_LEFT;
	r.topLeft.y = PBAR_TOP;
	r.extent.x = PBAR_WIDTH;
	r.extent.y = PBAR_HEIGHT;
	
	WinDrawRectangleFrame(rectangleFrame, &r);
	RctInsetRectangle(&r, 1);
	WinEraseRectangleFrame(rectangleFrame, &r);
	RctCopyRectangle(&r, &e);
	r.extent.x = (UInt32)num * r.extent.x / of;
	e.extent.x -= r.extent.x;
	e.topLeft.x += r.extent.x;
	WinDrawRectangle(&r, 0);
	WinEraseRectangle(&e, 0);
}

static void uiPrvErrAlert(Err e)
{
	if (e == errNone)
		SysFatalAlert(mStrTaskFailedSuccessfully);
	else {
	
		UInt8 specifier = e & 0xff;
		MemHandle mh;
		char x[9];
		
		if (specifier) {
		
			mh = DmGetResource('err!', 0x1000 + (e >> 8));
			if (mh) {
				
				char *res = MemHandleLock(mh), *end = res + MemHandleSize(mh), *cls = res;
				
				while (res < end && specifier--)
					res += StrLen(res) + 1;
				
				if (res < end) {
					
					FrmCustomAlert(1000, mStrAnErrorOccured, cls, res);
					MemHandleUnlock(mh);
					return;
				}
				
				MemHandleUnlock(mh);
			}
		}
		
		StrIToH(x, e);
		FrmCustomAlert(1000, mStrError0x, x + 4, mStrOccured);
	}
}

static void uiPrvUpdateProgressF(void *userData, UInt16 blocksDone, UInt16 blocksOf)
{
	char x[32];
	UInt16 percent, len;
	
	if (!blocksDone)
		FrmAlert(1002);	//plug and unplug now...

	percent = (blocksDone * 100UL + blocksOf / 2) / blocksOf;

	StrCopy(x, "      ");
	StrIToA(x + StrLen(x), percent);
	StrCat(x,"%      ");
	len = StrLen(x);
	WinDrawChars(x, len, PBAR_LEFT + (PBAR_WIDTH - FntCharsWidth(x, len)) / 2, PBAR_TOP - 15);

	uiPrvPbarDraw(blocksDone, blocksOf);
	EvtResetAutoOffTimer();
}

static void doFlash(void)
{
	MemHandle mh = DmGetResource('fwre', 0);
	FormPtr fp;
	Err e;
	
	fp = FrmInitForm(1000);
	FrmDrawForm(fp);
	FrmSetActiveForm(fp);
	
	e = updatePerform(MemHandleLock(mh), MemHandleSize(mh), uiPrvUpdateProgressF, NULL);
	MemHandleUnlock(mh);
	DmReleaseResource(mh);
	
	if (e != errNone)
		uiPrvErrAlert(e);
	
	FrmSetActiveForm(NULL);
	FrmEraseForm(fp);
	FrmDeleteForm(fp);
	
	FrmAlert(1001);
	SysReset();
}

UInt32 __Startup__(void);
UInt32 __Startup__(void)
{
	void *prevGlobalsP, *globalsP;
	SysAppInfoPtr appInfoP;
	
	if (SysAppStartup(&appInfoP, &prevGlobalsP, &globalsP) == errNone) {
		
		if (appInfoP->cmd == sysAppLaunchCmdNormalLaunch)
			doFlash();
			
		SysAppExit(appInfoP, prevGlobalsP, globalsP);
	}
	
	return 0;
}














